import numpy as np
from numba import prange

from ...geometry._algorithms.bounds import total_bounds_interleaved
from ...geometry._algorithms.orientation import triangle_orientation
from ...utils import ngjit, ngpjit


@ngjit
def segment_intersects_point(ax0, ay0, ax1, ay1, bx, by):
    """
    Test whether a 2-dimensional line segment intersects with a point

    Args:
        ax0, ay0: coordinates of start of segment
        ax1, ay1: coordinates of end of segment
        bx, by: coordinates of point

    Returns:
        True if segment intersects point, False otherwise
    """
    # Check bounds
    if bx < min(ax0, ax1) or bx > max(ax0, ax1):
        return False
    if by < min(ay0, ay1) or by > max(ay0, ay1):
        return False

    # Use cross product to test whether point is exactly on line
    # S is vector from segment start to segment end
    sx = ax1 - ax0
    sy = ay1 - ay0

    # P is vector from segment start to point
    px = bx - ax0
    py = by - ay0

    # Compute cross produce of S and P
    sxp = sx * py - sy * px

    return sxp == 0


@ngjit
def segments_intersect_1d(ax0, ax1, bx0, bx1):
    """
    Test whether two 1-dimensional line segments overlap
    Args:
        ax0, ax1: coords of endpoints of first segment
        bx0, bx1: coords of endpoints of second segment

    Returns:
        True if segments overlap, False otherwise
    """
    # swap inputs so that *x1 >= *x0
    if ax1 < ax0:
        ax0, ax1 = ax1, ax0
    if bx1 < bx0:
        bx0, bx1 = bx1, bx0

    return max(ax0, bx0) <= min(ax1, bx1)


@ngjit
def segments_intersect(ax0, ay0, ax1, ay1, bx0, by0, bx1, by1):
    """
    Test whether two 2-dimensional line segments intersect

    Args:
        ax0, ay0: coordinates of start of first segment
        ax1, ay1: coordinates of end of first segment
        bx0, by0: coordinates of start of second segment
        bx1, by1: coordinates of end of second segment

    Returns:
        True if segments intersect, False otherwise
    """
    if not segments_intersect_1d(ax0, ax1, bx0, bx1):
        # x projection of segments do not intersect, segments do not cross
        return False
    if not segments_intersect_1d(ay0, ay1, by0, by1):
        # y projection of segments do not intersect, segments do not cross
        return False

    a_zero = ax0 == ax1 and ay0 == ay1
    b_zero = bx0 == bx1 and by0 == by1
    if a_zero and not b_zero and (
            ax0 == bx0 and ay0 == by0 or ax0 == bx1 and ay0 == by1
    ):
        # a is zero length line that is identical to an end point of b
        return True
    elif b_zero and not a_zero and (
            bx0 == ax0 and by0 == ay0 or bx0 == ax1 and by0 == ay1
    ):
        # a is zero length line that is identical to an end point of b
        return True
    elif a_zero or b_zero:
        # a or b is zero length and does not match a vertex of the other line
        return False

    b0_orientation = triangle_orientation(ax0, ay0, ax1, ay1, bx0, by0)
    b1_orientation = triangle_orientation(ax0, ay0, ax1, ay1, bx1, by1)
    if b0_orientation == 0 and b1_orientation == 0:
        # b0 and b1 lie on line from a0 to a1, segments are collinear and intersect
        return True
    elif b0_orientation == b1_orientation:
        # b0 and b1 lie on the same side of line from a0 to a1, segments do not
        # intersect
        return False

    a0_orientation = triangle_orientation(bx0, by0, bx1, by1, ax0, ay0)
    a1_orientation = triangle_orientation(bx0, by0, bx1, by1, ax1, ay1)
    if a0_orientation == 0 and a1_orientation == 0:
        # a0 and a1 lie on line from b0 to b1, segments are collinear and cross
        return True
    elif a0_orientation == a1_orientation:
        # a0 and a1 are on the same side of line from b0 to b1, segments do not cross
        return False

    return True


@ngjit
def point_intersects_polygon(x, y, values, value_offsets):
    """
    Test whether a point intersects with a polygon

    Args:
        x, y: coordinates of test point
        values: array of interleaved coordinates of the polygon and holes
        value_offsets: array of offsets into values that separate the rings that
            compose the polygon. The first ring is the outer shell and subsequent rings
            are holes contained in this shell.

    Returns:
        True if the test point intersects with the polygon, False otherwise
    """
    winding_number = 0
    for i in range(len(value_offsets) - 1):
        start = value_offsets[i]
        stop = value_offsets[i + 1]

        for k in range(start, stop - 2, 2):
            x0 = values[k]
            y0 = values[k + 1]
            x1 = values[k + 2]
            y1 = values[k + 3]

            if y1 == y0:
                # skip horizontal edges
                continue

            # Make sure the y1 > y0 and keep track of whether edge was
            # originally ascending vertically
            if y1 < y0:
                ascending = -1
                y0, y1 = y1, y0
                x0, x1 = x1, x0
            else:
                ascending = 1

            # Reject edges that are fully above, below, or to the left of test point
            if y0 >= y or y1 < y or (x0 < x and x1 < x):
                continue

            if x0 >= x and x1 >= x:
                # Edge is fully to the right of test point, so we know that a ray to
                # the right will intersect the edge
                winding_number += ascending
            else:
                # Check if edge is to the right of test point using cross product
                # A is vector from test point to lower vertex
                ax = x0 - x
                ay = y0 - y

                # B is vector from test point to upper vertex
                bx = x1 - x
                by = y1 - y

                # Compute cross produce of A and B
                axb = ax * by - ay * bx

                if axb > 0 or (axb == 0 and ascending):
                    # Edge intersects with ray
                    winding_number += ascending

    return winding_number != 0


@ngpjit
def multipoints_intersect_bounds(
        x0, y0, x1, y1, flat_values, start_offsets, stop_offsets, result
):
    """
    Test whether each multipoint in a collection of multipoints intersects with
    the supplied bounds

    Args:
        x0, y0, x1, y1: Bounds coordinates
        flat_values: Interleaved point coordinates
        start_offsets, stop_offsets:  start and stop offsets into flat_values
            separating individual multipoints
        result: boolean array to be provided by the caller into which intersection
            results will be written; must be at least as long as start_offsets

    Returns:
        None
    """
    # Initialize results
    n = len(start_offsets)
    result.fill(False)

    # Orient rectangle
    if x1 < x0:
        x0, x1 = x1, x0
    if y1 < y0:
        y0, y1, = y1, y0

    for i in prange(n):
        start = start_offsets[i]
        stop = stop_offsets[i]

        # Check for points in rect
        point_in_rect = False
        for j in range(start, stop, 2):
            x = flat_values[j]
            y = flat_values[j + 1]
            if x0 <= x and x <= x1 and y0 <= y and y <= y1:
                point_in_rect = True
                break

        if point_in_rect:
            result[i] = True
            continue

    return result


@ngjit
def _perform_line_intersect_bounds(
        i, x0, y0, x1, y1, flat_values, start_offsets, stop_offsets, result
):
    start = start_offsets[i]
    stop = stop_offsets[i]

    # compute bounding box for line
    bounds = total_bounds_interleaved(flat_values[start:stop])

    if bounds[0] > x1 or bounds[1] > y1 or bounds[2] < x0 or bounds[3] < y0:
        # bounds outside of rect, does not intersect
        return

    if (bounds[0] >= x0 and bounds[2] <= x1 or
            bounds[1] >= y0 and bounds[3] <= y1):
        # bounds is fully contained in rect when both are projected onto the
        # x or y axis
        result[i] = True
        return

    # Check for vertices in rect
    vert_in_rect = False
    for j in range(start, stop, 2):
        x = flat_values[j]
        y = flat_values[j + 1]
        if x0 <= x and x <= x1 and y0 <= y and y <= y1:
            vert_in_rect = True
            break

    if vert_in_rect:
        result[i] = True
        return

    # Check for segment that crosses rectangle edge
    segment_intersects = False
    for j in range(start, stop - 2, 2):
        ex0 = flat_values[j]
        ey0 = flat_values[j + 1]
        ex1 = flat_values[j + 2]
        ey1 = flat_values[j + 3]

        # top
        if segments_intersect(ex0, ey0, ex1, ey1, x0, y1, x1, y1):
            segment_intersects = True
            break

        # bottom
        if segments_intersect(ex0, ey0, ex1, ey1, x0, y0, x1, y0):
            segment_intersects = True
            break

        # left
        if segments_intersect(ex0, ey0, ex1, ey1, x0, y0, x0, y1):
            segment_intersects = True
            break

        # right
        if segments_intersect(ex0, ey0, ex1, ey1, x1, y0, x1, y1):
            segment_intersects = True
            break

    if segment_intersects:
        result[i] = True


@ngjit
def lines_intersect_bounds(
        x0, y0, x1, y1, flat_values, start_offsets, stop_offsets, result
):
    """
    Test whether each line in a collection of lines intersects with the supplied bounds

    Args:
        x0, y0, x1, y1: Bounds coordinates
        flat_values: Interleaved line coordinates
        start_offsets, stop_offsets:  start and stop offsets into flat_values
            separating individual lines
        result: boolean array to be provided by the caller into which intersection
            results will be written; must be at least as long as start_offsets

    Returns:
        None
    """
    # Initialize results
    n = len(start_offsets)
    result.fill(False)

    # Orient rectangle
    if x1 < x0:
        x0, x1 = x1, x0
    if y1 < y0:
        y0, y1, = y1, y0

    if x0 == x1 or y0 == y1:
        # Zero width/height rect does not intersect with anything
        return

    for i in range(n):
        _perform_line_intersect_bounds(
            i, x0, y0, x1, y1, flat_values, start_offsets, stop_offsets, result
        )

    return result


@ngjit
def multilines_intersect_bounds(
        x0, y0, x1, y1, flat_values, start_offsets0, stop_offsets0, offsets1, result
):
    """
    Test whether each multiline in a collection of multilines intersects with the
    supplied bounds

    Args:
        x0, y0, x1, y1: Bounds coordinates
        flat_values: Interleaved line coordinates
        start_offsets0, stop_offsets0:  start and stop offsets into offsets1
            separating individual multilines
        offsets1: Offsets into flat_values separating individual lines
        result: boolean array to be provided by the caller into which intersection
            results will be written; must be at least as long as start_offsets

    Returns:
        None
    """
    # Initialize results
    n = len(start_offsets0)
    result.fill(False)

    if len(start_offsets0) < 1:
        # Empty array
        return

    # Orient rectangle
    if x1 < x0:
        x0, x1 = x1, x0
    if y1 < y0:
        y0, y1, = y1, y0

    if x0 == x1 or y0 == y1:
        # Zero width/height rect does not intersect with anything
        return

    # Populate results
    for i in range(n):
        # Numba has issues with following line when jit(parallel=True) is used:
        # Invalid use of Function(<intrinsic wrap_index>) with argument(s) of type(s):
        #   (uint32, int64)
        element_offsets = offsets1[start_offsets0[i]:stop_offsets0[i] + 1]
        num_lines = len(element_offsets) - 1
        element_result = np.zeros(num_lines, dtype=np.bool_)
        for j in range(num_lines):
            _perform_line_intersect_bounds(
                j, x0, y0, x1, y1,
                flat_values, element_offsets[:-1], element_offsets[1:], element_result
            )
        result[i] = element_result.any()


@ngjit
def _perform_polygon_intersect_bounds(i, x0, y0, x1, y1, flat_values, start_offsets0,
                                      stop_offsets0, offsets1, result):
    start0 = start_offsets0[i]
    stop0 = stop_offsets0[i]
    start1 = offsets1[start0]
    stop1 = offsets1[stop0]

    # compute bounding box for polygon.
    bounds = total_bounds_interleaved(flat_values[start1:stop1])

    if bounds[0] > x1 or bounds[1] > y1 or bounds[2] < x0 or bounds[3] < y0:
        # bounds outside of rect, does not intersect
        return

    if (bounds[0] >= x0 and bounds[2] <= x1 or
            bounds[1] >= y0 and bounds[3] <= y1):
        # bounds is fully contained in rect when both are projected onto the
        # x or y axis
        result[i] = True
        return

        # Check for vertices in rect
    vert_in_rect = False
    for k in range(start1, stop1, 2):
        x = flat_values[k]
        y = flat_values[k + 1]
        if x0 <= x and x <= x1 and y0 <= y and y <= y1:
            vert_in_rect = True
            break

    if vert_in_rect:
        result[i] = True
        return

    # Check for segment that crosses rectangle edge
    segment_intersects = False
    for j in range(start0, stop0):
        for k in range(offsets1[j], offsets1[j + 1] - 2, 2):
            ex0 = flat_values[k]
            ey0 = flat_values[k + 1]
            ex1 = flat_values[k + 2]
            ey1 = flat_values[k + 3]

            # top
            if segments_intersect(ex0, ey0, ex1, ey1, x0, y1, x1, y1):
                segment_intersects = True
                break

            # bottom
            if segments_intersect(ex0, ey0, ex1, ey1, x0, y0, x1, y0):
                segment_intersects = True
                break

            # left
            if segments_intersect(ex0, ey0, ex1, ey1, x0, y0, x0, y1):
                segment_intersects = True
                break

            # right
            if segments_intersect(ex0, ey0, ex1, ey1, x1, y0, x1, y1):
                segment_intersects = True
                break

        if segment_intersects:
            result[i] = True

    if segment_intersects:
        return

        # Check if a rectangle corners is in rect
    polygon_offsets = offsets1[start0:stop0 + 1]
    if point_intersects_polygon(x0, y0, flat_values, polygon_offsets):
        result[i] = True
        return
    if point_intersects_polygon(x1, y0, flat_values, polygon_offsets):
        result[i] = True
        return
    if point_intersects_polygon(x1, y1, flat_values, polygon_offsets):
        result[i] = True
        return
    if point_intersects_polygon(x0, y1, flat_values, polygon_offsets):
        result[i] = True
        return


@ngjit
def polygons_intersect_bounds(
        x0, y0, x1, y1, flat_values, start_offsets0, stop_offsets0, offsets1, result
):
    """
    Test whether each polygon in a collection of polygons intersects with the
    supplied bounds

    Args:
        x0, y0, x1, y1: Bounds coordinates
        flat_values: Interleaved vertex coordinates
        start_offsets0, stop_offsets0: start and stop offsets into offsets1 separating
            individual polygons
        offsets1:  offsets into flat_values separating individual polygons rings
        result: boolean array to be provided by the caller into which intersection
            results will be written; must be at least as long as start_offsets

    Returns:
        None
    """
    # Initialize results
    n = len(start_offsets0)
    result.fill(False)

    # Orient rectangle
    if x1 < x0:
        x0, x1 = x1, x0
    if y1 < y0:
        y0, y1, = y1, y0

    if x0 == x1 or y0 == y1:
        # Zero width/height rect does not intersect with anything
        return

    for i in range(n):
        _perform_polygon_intersect_bounds(
            i, x0, y0, x1, y1, flat_values,
            start_offsets0, stop_offsets0, offsets1, result
        )

    return result


@ngjit
def multipolygons_intersect_bounds(
        x0, y0, x1, y1, flat_values,
        start_offsets0, stop_offsets0, offsets1, offsets2, result
):
    """
    Test whether each multipolygon in a collection of multipolygons intersects with
    the supplied bounds

    Args:
        x0, y0, x1, y1: Bounds coordinates
        flat_values: Interleaved vertex coordinates
        start_offsets0, stop_offsets0: start and stop offsets into offsets1 separating
            individual multipolygons
        offsets1: offsets into offsets2 separating individual polygons
        offsets2: offsets into flat_values separating individual polygon rings
        result: boolean array to be provided by the caller into which intersection
            results will be written; must be at least as long as start_offsets

    Returns:
        None
    """

    # Initialize results
    n = len(start_offsets0)
    result.fill(False)

    if len(start_offsets0) < 1:
        # Empty array
        return

    # Orient rectangle
    if x1 < x0:
        x0, x1 = x1, x0
    if y1 < y0:
        y0, y1, = y1, y0

    if x0 == x1 or y0 == y1:
        # Zero width/height rect does not intersect with anything
        return

    # Populate results
    for i in range(n):
        polygon_offsets = offsets1[start_offsets0[i]:stop_offsets0[i] + 1]
        num_polys = len(polygon_offsets) - 1
        element_result = np.zeros(num_polys, dtype=np.bool_)
        for j in range(num_polys):
            _perform_polygon_intersect_bounds(
                j, x0, y0, x1, y1, flat_values,
                polygon_offsets[:-1], polygon_offsets[1:], offsets2, element_result
            )

        result[i] = element_result.any()



import pickle

import hypothesis.strategies as st
import numpy as np
import pytest
from hypothesis import given, settings
from hypothesis.extra.numpy import arrays

from spatialpandas.spatialindex import HilbertRtree

# ### hypothesis settings ###
hyp_settings = settings(deadline=None)


# ### Custom strategies ###

@st.composite
def st_bounds(draw, n_min=1, n_max=3):
    n = draw(st.integers(min_value=n_min, max_value=n_max))
    dim_starts = [draw(st.floats(0, 10**(i + 1) - 1)) for i in range(n)]
    dim_widths = [draw(st.floats(0, 10 ** i)) for i in range(n)]
    dim_ends = [s + w for s, w in zip(dim_starts, dim_widths)]
    return tuple(dim_starts + dim_ends)


@st.composite
def st_bounds_array(draw, n_min=1, n_max=3, min_size=1, max_size=1000):
    n = draw(st.integers(min_value=n_min, max_value=n_max))
    size = draw(st.integers(min_value=min_size, max_value=max_size))

    dim_starts = draw(
        arrays(elements=st.floats(0, 9), shape=(size, n), dtype='float64')
    )
    dim_widths = draw(
        arrays(elements=st.floats(0, 1), shape=(size, n), dtype='float64')
    )
    dim_ends = dim_starts + dim_widths
    bounds_array = np.concatenate([dim_starts, dim_ends], axis=1)
    if bounds_array.size == 0:
        bounds_array = bounds_array.reshape(0, 2*n)
    return bounds_array


st_page_size = st.integers(min_value=-1, max_value=1000000)

# ### Test custom strategies ###
@given(st_bounds())
def test_bounds_generation(bounds):
    assert len(bounds) % 2 == 0
    n = len(bounds) // 2
    for d in range(n):
        assert bounds[d + n] >= bounds[d]


@given(st_bounds_array())
def test_bounds_array_generation(bounds_array):
    assert bounds_array.shape[1] % 2 == 0
    n = bounds_array.shape[1] // 2
    for d in range(n):
        assert np.all(bounds_array[:, d + n] >= bounds_array[:, d])


# ### utilities ###
def intersects_bruteforce(query_bounds, bounds):
    if len(bounds) == 0:
        return set()

    outside_mask = np.zeros(bounds.shape[0], dtype=np.bool_)
    n = bounds.shape[1] // 2
    for d in range(n):
        outside_mask |= (bounds[:, d + n] < query_bounds[d])
        outside_mask |= (bounds[:, d] > query_bounds[d + n])

    return set(np.nonzero(~outside_mask)[0])


def covers_bruteforce(query_bounds, bounds):
    if len(bounds) == 0:
        return set()

    not_covers_mask = np.zeros(bounds.shape[0], dtype=np.bool_)
    n = bounds.shape[1] // 2
    for d in range(n):
        not_covers_mask |= (bounds[:, d] < query_bounds[d])
        not_covers_mask |= (bounds[:, d + n] > query_bounds[d + n])

    return set(np.nonzero(~not_covers_mask)[0])


# ### Hypothesis tests ###
@given(st_bounds_array(), st_page_size)
@hyp_settings
def test_rtree_intersects_input_bounds(bounds_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)

    # query with bounds array, these were the input
    for i in range(bounds_array.shape[0]):
        b = bounds_array[i, :]
        intersected = set(rt.intersects(np.array(b)))

        # Should at least select itself
        assert i in intersected

        # Compare to brute force implementation
        assert intersected == intersects_bruteforce(b, bounds_array)


@given(st_bounds_array(min_size=0, max_size=0), st_page_size)
@hyp_settings
def test_rtree_empty(bounds_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)
    n = bounds_array.shape[1] // 2
    query_bounds = (-np.inf,) * n + (np.inf,) * n
    assert set(rt.intersects(query_bounds)) == set()
    covers, overlaps = rt.covers_overlaps(query_bounds)
    assert set(covers) == set()
    assert set(overlaps) == set()
    assert rt.total_bounds == (np.nan,) * (2 * n)


@given(
    st_bounds_array(n_min=2, n_max=2),
    st_bounds_array(n_min=2, n_max=2, min_size=10),
    st_page_size
)
@hyp_settings
def test_rtree_intersects_different_bounds_2d(bounds_array, query_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)

    # query with query array
    for i in range(query_array.shape[0]):
        b = query_array[i, :]
        intersected = set(rt.intersects(np.array(b)))

        # Compare to brute force implementation
        assert intersected == intersects_bruteforce(b, bounds_array)


@given(st_bounds_array(n_min=1), st_page_size)
@hyp_settings
def test_rtree_intersects_all(bounds_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)

    # query with query array
    n = bounds_array.shape[1] // 2
    b_mins = bounds_array[:, :n].min(axis=0)
    b_maxes = bounds_array[:, n:].max(axis=0)
    b = np.concatenate([b_mins, b_maxes])
    intersected = set(rt.intersects(np.array(b)))

    # All indices should be selected
    assert intersected == set(range(bounds_array.shape[0]))


@given(st_bounds_array(), st_page_size)
@hyp_settings
def test_rtree_intersects_none(bounds_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)

    # query with query array
    n = bounds_array.shape[1] // 2
    b = ([11] * n) + ([100] * n)
    intersected = set(rt.intersects(np.array(b)))

    # Nothing should be selected
    assert intersected == set()


@pytest.mark.slow
@given(st_bounds_array(), st_page_size)
@hyp_settings
def test_rtree_covers_overlaps_input_bounds(bounds_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)

    # query with bounds array, these were the input
    for i in range(bounds_array.shape[0]):
        b = bounds_array[i, :]
        covers, overlaps = rt.covers_overlaps(np.array(b))
        covers = set(covers)
        overlaps = set(overlaps)

        # Should have nothing in common
        assert covers.isdisjoint(overlaps)

        # covered should contain all bounding boxes that are fully covered by query
        all_covers = covers_bruteforce(b, bounds_array)
        if covers != all_covers:
            rt.covers_overlaps(np.array(b))
        assert covers == all_covers

        # covered and overlaps together should contain all intersecting bounding
        # boxes (overlaps will typically contain others as well)
        intersects = intersects_bruteforce(b, bounds_array)
        assert intersects == covers.union(overlaps)


@given(st_bounds_array(), st_page_size)
@hyp_settings
def test_rtree_pickle(bounds_array, page_size):
    # Build rtree
    rt = HilbertRtree(bounds_array, page_size=page_size)

    # Serialize uninitialized rtree
    rt2 = pickle.loads(pickle.dumps(rt))
    if bounds_array.size == 0:
        assert isinstance(rt2, HilbertRtree)
        return

    rt2_result = set(rt2.intersects(bounds_array[0, :]))

    # Call intersection to construct numba rtree
    rt_result = set(rt.intersects(bounds_array[0, :]))

    # Serialize initialized rtree
    rt3 = pickle.loads(pickle.dumps(rt))
    rt3_result = set(rt3.intersects(bounds_array[0, :]))
    assert rt_result == rt2_result and rt_result == rt3_result



import numpy as np

from ...utils import ngjit


@ngjit
def total_bounds_interleaved(values):
    """
    compute bounds
    """
    xmin = np.inf
    ymin = np.inf
    xmax = -np.inf
    ymax = -np.inf

    for i in range(0, len(values), 2):
        x = values[i]
        if np.isfinite(x):
            xmin = min(xmin, x)
            xmax = max(xmax, x)

        y = values[i + 1]
        if np.isfinite(y):
            ymin = min(ymin, y)
            ymax = max(ymax, y)

    # replace infinite values with nan in the case no finite values were found
    if not np.isfinite(xmin):
        xmin = xmax = np.nan
    if not np.isfinite(ymin):
        ymin = ymax = np.nan

    return (xmin, ymin, xmax, ymax)


@ngjit
def total_bounds_interleaved_1d(values, offset):
    """
    compute bounds
    """
    vmin = np.inf
    vmax = -np.inf

    for i in range(0, len(values), 2):
        v = values[i + offset]
        if np.isfinite(v):
            vmin = min(vmin, v)
            vmax = max(vmax, v)

    if np.isfinite(vmin):
        return (vmin, vmax)
    else:
        return (np.nan, np.nan)


@ngjit
def bounds_interleaved(flat_values, flat_value_offsets):
    # Initialize bounds array that will be populated and returned
    bounds = np.full((len(flat_value_offsets) - 1, 4), np.nan, dtype=np.float64)

    for i in range(len(flat_value_offsets) - 1):
        start = flat_value_offsets[i]
        stop = flat_value_offsets[i + 1]
        bounds[i, :] = total_bounds_interleaved(flat_values[start:stop])

    return bounds



import param as _param

from . import geometry, spatialindex, tools
from .geodataframe import GeoDataFrame
from .geoseries import GeoSeries
from .tools.sjoin import sjoin

try:
    import dask.dataframe  # noqa

    # Import to trigger registration of types with Dask
    import spatialpandas.dask  # noqa
except ImportError:
    # Dask dataframe not available
    pass

__version__ = str(
    _param.version.Version(
        fpath=__file__,
        archive_commit="$Format:%h$",
        reponame="spatialpandas",
    ))

__all__ = [
    "GeoDataFrame",
    "GeoSeries",
    "geometry",
    "sjoin",
    "spatialindex",
    "tools",
]



try:
    import shapely.geometry as sg
except ImportError:
    sg = None

try:
    import geopandas as gp
except Exception:
    # Catch broad exception to handle the case of a corrupt geopandas installation
    gp = None



import numpy as np
import pyarrow as pa
from dask.dataframe.extensions import make_array_nonempty
from pandas.core.dtypes.dtypes import register_extension_dtype

from ..geometry import Polygon
from ..geometry._algorithms.intersection import multipolygons_intersect_bounds
from ..geometry._algorithms.measures import compute_area, compute_line_length
from ..geometry._algorithms.orientation import orient_polygons
from ..geometry.base import GeometryDtype
from ..geometry.baselist import (
    GeometryList,
    GeometryListArray,
    _geometry_map_nested3,
)
from ..geometry.multiline import MultiLine, MultiLineArray


@register_extension_dtype
class MultiPolygonDtype(GeometryDtype):
    _geometry_name = 'multipolygon'

    @classmethod
    def construct_array_type(cls, *args):
        if len(args) > 0:
            raise NotImplementedError("construct_array_type does not support arguments")
        return MultiPolygonArray


class MultiPolygon(GeometryList):
    _nesting_levels = 2

    @classmethod
    def construct_array_type(cls):
        return MultiPolygonArray

    @classmethod
    def _shapely_to_coordinates(cls, shape):
        import shapely.geometry as sg
        if isinstance(shape, sg.MultiPolygon):
            multipolygon = []
            for polygon in shape.geoms:
                polygon_coords = Polygon._shapely_to_coordinates(polygon)
                multipolygon.append(polygon_coords)

            return multipolygon
        elif isinstance(shape, sg.Polygon):
            return [Polygon._shapely_to_coordinates(shape)]
        else:
            raise ValueError("""
Received invalid value of type {typ}. Must be an instance of Polygon or MultiPolygon
""".format(typ=type(shape).__name__))

    def to_shapely(self):
        """
        Convert to shapely shape

        Returns:
            shapely MultiPolygon shape
        """
        import shapely.geometry as sg
        polygon_arrays = self.data.as_py()

        polygons = []
        for polygon_array in polygon_arrays:
            ring_arrays = [np.array(line_coords).reshape(len(line_coords) // 2, 2)
                           for line_coords in polygon_array]

            rings = [sg.LinearRing(ring_array) for ring_array in ring_arrays]
            polygons.append(sg.Polygon(shell=rings[0], holes=rings[1:]))

        return sg.MultiPolygon(polygons=polygons)

    @classmethod
    def from_shapely(cls, shape, orient=True):
        """
        Build a spatialpandas MultiPolygon object from a shapely shape

        Args:
            shape: A shapely Polygon or MultiPolygon shape
            orient: If True (default), reorder polygon vertices so that outer shells
                    are stored in counter clockwise order and holes are stored in
                    clockwise order.  If False, accept vertices as given. Note that
                    while there is a performance cost associated with this operation
                    some algorithms will not behave properly if the above ordering
                    convention is not followed, so only set orient=False if it is
                    known that this convention is followed in the input data.
        Returns:
            spatialpandas MultiPolygon
        """
        import shapely.geometry as sg
        if orient:
            if isinstance(shape, sg.Polygon):
                shape = sg.polygon.orient(shape)
            elif isinstance(shape, sg.MultiPolygon):
                shape = sg.MultiPolygon([sg.polygon.orient(poly) for poly in shape.geoms])

        shape_parts = cls._shapely_to_coordinates(shape)
        return cls(shape_parts)

    @property
    def boundary(self):
        new_offsets = self.buffer_offsets[1]
        new_data = pa.ListArray.from_arrays(new_offsets, self.buffer_values)
        return MultiLine(new_data)

    @property
    def length(self):
        return compute_line_length(self.buffer_values, self.buffer_inner_offsets)

    @property
    def area(self):
        return compute_area(self.buffer_values, self.buffer_inner_offsets)

    def intersects_bounds(self, bounds):
        x0, y0, x1, y1 = bounds
        result = np.zeros(1, dtype=np.bool_)
        offsets1, offsets2 = self.buffer_offsets
        offsets0 = np.array([0, len(offsets1) - 1], dtype=np.uint32)
        multipolygons_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1), self.buffer_values,
            offsets0[:-1], offsets0[1:], offsets1, offsets2, result
        )
        return result[0]


class MultiPolygonArray(GeometryListArray):
    _element_type = MultiPolygon
    _nesting_levels = 3

    @property
    def _dtype_class(self):
        return MultiPolygonDtype

    @classmethod
    def from_geopandas(cls, ga, orient=True):
        """
        Build a spatialpandas MultiPolygonArray from a geopandas GeometryArray or
        GeoSeries.

        Args:
            ga: A geopandas GeometryArray or GeoSeries of MultiPolygon or
                Polygon shapes.
            orient: If True (default), reorder polygon vertices so that outer shells
                    are stored in counter clockwise order and holes are stored in
                    clockwise order.  If False, accept vertices as given. Note that
                    while there is a performance cost associated with this operation
                    some algorithms will not behave properly if the above ordering
                    convention is not followed, so only set orient=False if it is
                    known that this convention is followed in the input data.

        Returns:
            MultiPolygonArray
        """
        mpa = super().from_geopandas(ga)
        if orient:
            return mpa.oriented()
        else:
            return mpa

    def oriented(self):
        missing = np.concatenate([self.isna(), [False]])
        buffer_values = self.buffer_values.copy()
        multipoly_offsets, poly_offsets, ring_offsets = self.buffer_offsets

        orient_polygons(buffer_values, poly_offsets, ring_offsets)

        pa_rings = pa.ListArray.from_arrays(
            pa.array(ring_offsets), pa.array(buffer_values)
        )
        pa_polys = pa.ListArray.from_arrays(
            pa.array(poly_offsets), pa_rings,
        )
        pa_multipolys = pa.ListArray.from_arrays(
            pa.array(multipoly_offsets, mask=missing), pa_polys
        )
        return self.__class__(pa_multipolys)

    @property
    def boundary(self):
        offsets = self.buffer_offsets
        inner_data = pa.ListArray.from_arrays(offsets[2], self.buffer_values)
        new_data = pa.ListArray.from_arrays(offsets[1][offsets[0]], inner_data)
        return MultiLineArray(new_data)

    @property
    def length(self):
        result = np.full(len(self), np.nan, dtype=np.float64)
        _geometry_map_nested3(
            compute_line_length,
            result,
            self.buffer_values,
            self.buffer_offsets,
            self.isna(),
        )
        return result

    @property
    def area(self):
        result = np.full(len(self), np.nan, dtype=np.float64)
        _geometry_map_nested3(
            compute_area,
            result,
            self.buffer_values,
            self.buffer_offsets,
            self.isna(),
        )
        return result

    def intersects_bounds(self, bounds, inds=None):
        x0, y0, x1, y1 = bounds
        offsets0, offsets1, offsets2 = self.buffer_offsets
        start_offsets0 = offsets0[:-1]
        stop_offsets0 = offsets0[1:]
        if inds is not None:
            start_offsets0 = start_offsets0[inds]
            stop_offsets0 = stop_offsets0[inds]

        result = np.zeros(len(start_offsets0), dtype=np.bool_)
        multipolygons_intersect_bounds(
            float(x0), float(y0), float(x1), float(y1), self.buffer_values,
            start_offsets0, stop_offsets0, offsets1, offsets2, result
        )
        return result


def _multi_polygon_array_non_empty(dtype):
    """
    Create an example length 2 array to register with Dask.
    See https://docs.dask.org/en/latest/dataframe-extend.html#extension-arrays
    """
    return MultiPolygonArray([
        [
            [[1.0, 1.0, 2.0, 1.0, 2.0, 2.0, 1.0, 2.0, 1.0, 1.0],
             [1.1, 1.1, 1.5, 1.9, 1.9, 1.1, 1.1, 1.1]]
        ],
        [
            [[0.0, 0.0, 1.0, 0.0, 2.0, 1.0, 0.5, 3.0, -1.0, 1.0, 0.0, 0.0],
             [0.2, 0.2, 0.5, 1.0, 0.8, 0.2, 0.2, 0.2],
             [0.5, 1.25, 0.3, 2.0, 0.8, 2.0, 0.5, 1.25]]
        ]
    ], dtype=dtype)


if make_array_nonempty:
    make_array_nonempty.register(MultiPolygonDtype)(_multi_polygon_array_non_empty)
